function [nodes,un1]=fem_1d_heat_solver(geom,problem_data,parameters)
% fem_1d_solver: solve du/dt-mu d^2u/dx^2+sigma u=f 
%                in Omega x (t0,T)
% with Dirichlet and/or Neumann boundary conditions
% by either P1-fem or P2-fem on a uniform grid.
%
% [nodes,un1]=fem_1d_heat_solver(geom,problem_data,parameters)
%
% Input: geom: struct with fields:
%          geom.xa, geom.xb = end-points of Omega
%          geom.t0, geom.tf = end-points of (t0,T)
%        problem_data: struct with coefficients and functions data
%              .mu =  mu (positive constant)
%              .sigma =  sigma (positive constant)
%              .f =f function handle @(x,t)
%              .gd0=@(t) Dirichlet datum at left end-point
%              .gd1=@(t) Dirichlet datum at right end-point
%              .gn0=@(t) Neumann datum at left end-point
%              .gn1=@(t) Neumann datum at right end-point
%              .u0=@(x) initial condition
%              .bc= 'xx' boundary conditions type:
%                  .bc(1)='d' if Dirichlet at left
%                  .bc(1)='n' if Neumann at left
%                  .bc(2)='d' if Dirichlet at right
%                  .bc(2)='n' if Neumann at right
%        parameters: struct with discretization parameters
%          .p = local polynomial degree (1 or 2)
%          .Ne =  number of elements of the partition
%          .theta = theta
%          .Nt = number of time steps
%          .plot_solution = 1 plot the solution at each time step
%                           0 do not plot the solution
%          .plot_axis=[umin,umax]; % umin and umax for the plot 
% (optional) .degree = degree of exactness of quadrature formulas
%                  (default value =4)
%
% Output: nodes = column array with the nodes of the mesh
%         un1 = column array of the numerical solution at the final time T

%   Written by Paola Gervasio
%   $Date: 2020/04/26$

if ~isfield(parameters,'degree')
  parameters.degree=4;
end

% assemble matrices
[mat,fem_space,tt]=fem_1d_heat_mat(geom,problem_data,parameters);


% initial condition at time t0
un=problem_data.u0(fem_space.nodes);
fn=[];
if parameters.plot_solution
  figure(1); clf
  [xh1,uh1]=fem_1d_evalsol(un,geom,parameters.p,parameters.Ne,20);
  pl=plot(xh1,uh1,'Linewidth',2);
  color=pl.Color;
  axis([xh1(1),xh1(end),parameters.plot_axis]);
  grid on; hold on
  xlabel('x'); ylabel('u_h^n');
  title(['t_n=',num2str(geom.t0)]);
  pause(0.1)
end

% loop on the time-step: knwon un, compute un1
for n=1:parameters.Nt
  
  % assemble the arrays:
  % fn1=f(t^(n+1))
  % g= (M/dt-(1-theta)A)u^n+(1-theta)f^n +theta f^(n+1)
  
  [fn1,g]=rhs_theta_method(un,mat,fem_space,parameters,problem_data,...
    tt(n),tt(n+1),fn);
  N=length(g); % size of the system

  % update Dirichlet boundary conditions
  if problem_data.bc(1)=='d'
    g(1)=problem_data.gd0(tt(n+1));  % left end-point
  end
  if problem_data.bc(2)=='d'
    g(N)=problem_data.gd1(tt(n+1)); % right end-point
  end
  
  % update the rhs to solve the system restricted to non-dirichlet dofs
  if ~isempty(fem_space.dof.dir)
    g0=g(fem_space.dof.nondir)-mat.Kb*g(fem_space.dof.dir);
  else
    g0=g;
  end
  
  % solve the linear system (mat.K0 is the Cholesky upper factor):
  % un10 is the dir-homogeneous solution
  y=mat.K0'\g0;
  un10=mat.K0\y;
  
  % assemble the solution: un1=un10+Rg (homoegenous solution + lifting of the
  % Dirichlet data)
  
  un1=zeros(N,1);
  % copy u0 in un1 (only nondirichlet dofs)
  un1(fem_space.dof.nondir)=un10;
  % copy Rg in un1 (only dirichlet dofs)
  if ~isempty(fem_space.dof.dir)
    un1(fem_space.dof.dir)=g(fem_space.dof.dir);
  end
  
  % plot the solution at this time-step
  if parameters.plot_solution
    [xh1,uh1]=fem_1d_evalsol(un1,geom,parameters.p,parameters.Ne,20);
    delete(pl);
    pl=plot(xh1,uh1,'Linewidth',2,'color',color);
    
    axis([xh1(1),xh1(end),parameters.plot_axis]);
    title(['t_n=',num2str(tt(n+1))]);
    pause 
  end
  % update un and fn for the next step
  un=un1;
  fn=fn1;
end

nodes=fem_space.nodes;
end

% ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
% fem-P1 and fem-P2: assemble the matrix and the rhs
% ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function [mat,fem_space,tt]=fem_1d_heat_mat(geom,problem_data,...
  parameters)
% Output: mat = struct with fields
%            .Kn = M/dt-(1-theta)*A
%            .K0 upper cholesky factor of K=M/dt+theta*A
%            .Kb block (nondir,dir) of K=M/dt+theta*A
%         fem_space =  struct with fields
%                 .nodes = mash in Omega
%                 .nov  = connectivity matrix
%                 .ne = number of elements in Omega;
%                 .basis = basis functions evaluated at the quadrature
%                 nodes
%                 .qf = struct with quadrature nodes and weights
%                 .dof  = struct with the lists of dirichlet and
%                 non-dirichlet dofs
%         tt = array of discrete times

p=parameters.p;
Ne=parameters.Ne;
Nt=parameters.Nt;
theta=parameters.theta;
degree=parameters.degree;

h=(geom.xb-geom.xa)/Ne; % mesh-size
N=p*Ne+1; % number of nodes
nodes=linspace(geom.xa,geom.xb,N)'; % nodes

dt=(geom.tf-geom.t0)/Nt;
tt=linspace(geom.t0,geom.tf,Nt+1)'; % time steps


% lists of dirichlet and non-dirichlet nodes
dof.dir=[];
dof.nondir=(1:N)';
if problem_data.bc(1)=='d'
  dof.dir=[dof.dir;1];
  dof.nondir=setdiff(dof.nondir,1) ;
end
if problem_data.bc(2)=='d'
  dof.dir=[dof.dir;N];
  dof.nondir=setdiff(dof.nondir,N) ;
end

nloc=p+1; % number of nodes inside each element
% connectivity matrix
nov=(1:nloc)'; nmax=nloc;
for ie=2:Ne
  nov=[nov,(nmax:nmax+p)'];
  nmax=nmax+p;
end

% mass matrix on the reference interval (0,1)
if p==1
  Ml=[ 2,1; 1,2 ]/6;
elseif p==2
  Ml=[4, 2, -1;
    2, 16, 2;
    -1, 2,  4]/30;
end

% assemble the global mass matrix
M=zeros(N);
for ie=1:Ne
  M(nov(:,ie),nov(:,ie))=M(nov(:,ie),nov(:,ie))+Ml*h;
end


%M=diag(sum(M,2)); % lumped mass matrix


% stiffness matrix on the reference interval (0,1)
if p==1
  Kl=[1,-1;-1,1];
elseif p==2
  Kl=[7, -8, 1;
    -8,16,-8;
    1,-8,7]/3;
end

% assemble the global stiffness matrix
K=zeros(N);
for ie=1:Ne
  K(nov(:,ie),nov(:,ie))=K(nov(:,ie),nov(:,ie))+Kl/h;
end


% matrix A
A=problem_data.mu*K+problem_data.sigma*M;
A=sparse(A);
M=sparse(M);

% matrix Kn=M/dt-(1-theta)*A
mat.Kn=M/dt-(1-theta)*A;

% matrix K=M/dt+theta*A
Kg=M/dt+theta*A;
mat.K0=chol(Kg(dof.nondir,dof.nondir));
mat.Kb=Kg(dof.nondir,dof.dir);
clear Kg;

% compute quadrature nodes and weights and
% evaluate the basis functions at the quad nodes
fem_space.nodes=nodes;
fem_space.nov=nov;
fem_space.ne=Ne;

[fem_space.basis,~,fem_space.qf]=fem_1d_setting(fem_space,p,degree);
fem_space.dof=dof;
end



function [fn1,g]=rhs_theta_method(un,mat,fem_space,parameters,...
  problem_data,tn,tn1,fn)
% rhs_theta_method:  assemble the arrays fn1=f(t^(n+1)) and
%  g= (M/dt-(1-theta)A)u^n+(1-theta)f^n +theta f^(n+1) for one step of 
%  the theta method.
% [fn1,g]=rhs_theta_method(un,mat,fem_space,parameters,...
%  problem_data,tn,tn1,fn)

N=length(un);
% if fn is empty, evaluate f^n
if isempty(fn)
  fn=zeros(N,1);
  for ie=1:parameters.Ne
    novl=fem_space.nov(:,ie);
    ff=problem_data.f(fem_space.qf.xq_ph(:,ie),tn);
    floc=fem_space.basis.phi'*(ff.*fem_space.qf.wq_ph(:,ie));
    fn(novl)=fn(novl)+floc;
  end
 % Neumann boundary conditions
  if problem_data.bc(1)=='n'
    fn(1)=fn(1)+problem_data.gn0(tn); % Neumann
  end
  if problem_data.bc(2)=='n'
  fn(N)=fn(N)+problem_data.gn1(tn); % Neumann
  end
end


% evaluate f^(n+1)
fn1=zeros(N,1);
for ie=1:parameters.Ne
  novl=fem_space.nov(:,ie);
  ff=problem_data.f(fem_space.qf.xq_ph(:,ie),tn1);
  floc=fem_space.basis.phi'*(ff.*fem_space.qf.wq_ph(:,ie));
  fn1(novl)=fn1(novl)+floc;
end
 % Neumann boundary conditions
  if problem_data.bc(1)=='n'
    fn1(1)=fn1(1)+problem_data.gn0(tn1); % Neumann
  end
  if problem_data.bc(2)=='n'
  fn1(N)=fn1(N)+problem_data.gn1(tn1); % Neumann
  end

g=mat.Kn*un+(1-parameters.theta)*fn+parameters.theta*fn1;

end

